PORT = 8086

MAX_CACHE_AGE = 600  # seconds

DEBUG = True

PROJECTS_FILE = 'projects.yaml'
# REPOS_CACHE_DIR = '/var/cache/webhaak'
REPOS_CACHE_DIR = '/tmp/webhaak'

SECRETKEY = 'blah123'

LOG_LOCATION = 'webhaak.log'
LOG_BACKUP_COUNT = 10

PUSHOVER_USERKEY = 'umuy2covsiv9irfr57on43smideaat'
PUSHOVER_APPTOKEN = 'ahfjrmhhwwf4p2ct7grn4jqap5j5ho'
